%--------------------------------------------------------------------------
% Name:            ExtendedKalmanFilterDemo.m
%
% Description:     Application of an extended Kalman filter to estimate
%                  position of a target object.
%
% Author:          Thor Audun Steen, (based on Brian Moore)
%
% Date:            June 14, 2014
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% Model Constants
%--------------------------------------------------------------------------
% Time sample
Ts = 1;

% Object Dynamics
G1 = objectDynamics([12 18],[1 2]);
G2 = objectDynamics([10 24],[1 4]);

sys = ss(G1);
[Am,Bm,Cm,Dm]=bilinear(sys.A,sys.B,sys.C,sys.D,1/Ts);

sys2 = ss(G2);
sysd = c2d(sys2,Ts,'zoh');



% Size of X(k)
nA = size(StateEquation(sysd,Ts,[1;1;1]),1);
nC = size(MeasurementEquation(nA),1);

% State-error Jacobian
W = diag([1.6 0.4 2 ones(1,15)]);

% Measurement-state Jacobian
H = MeasurementEquation(nA);

% Measurement-error Jacobian
V = diag([1.6 0.8 1]);
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% State variable initializations
%--------------------------------------------------------------------------
% Number of iterations  
N = 400;

% True State
x = zeros(nA,N);

% Apriori state estimates
x_apriori = zeros(nA,N);

% Aposteriori state estimates
x_aposteriori = zeros(nA,N);

% Apriori error covariance estimates
P_apriori = zeros(nA,nA,N);

% Aposteriori error covariance estimates
P_aposteriori = zeros(nA,nA,N);

% Measurements
z = zeros(nC,N);

% Kalman Gain
K = zeros(nA,nC,N);
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% Knobs to turn
%--------------------------------------------------------------------------
% True initial state
x(:,1) = [1; pi/180*100; 30; zeros(nA-3,1)];

% Initial aposteriori state estimate, X_k|k
x_aposteriori(:,1) = [0.5; pi/180*50; 20; zeros(nA-3,1)];

% Process noise covariance
Q = W*0.5*diag([0.005 0.005 0.5, 0.01*ones(1,15)])*W';

% Measurement noise covariance
R = V*10*diag([10 10 5])*V';

% Initial aposteriori error covariance estimate, P_k|k
P_aposteriori(:,:,1) = diag([1 0.1 1, 0.1*ones(1,15)]);
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% True State
%--------------------------------------------------------------------------
t = 0:Ts:N;

r_rho = [0.5*ones(1,100/Ts),1.5*ones(1,100/Ts),0.7*ones(1,100/Ts),2.6*ones(1,100/Ts+1)];
r_psi = pi/180*[100*ones(1,50/Ts),30*ones(1,200/Ts),60*ones(1,50/Ts),-30*ones(1,100/Ts+1)];
r_z   = [30*ones(1,130/Ts),50*ones(1,100/Ts),20*ones(1,70/Ts),80*ones(1,100/Ts+1)];

Xr = [r_rho; r_psi; r_z];

Xm = zeros(10,length(t));
Xy = zeros(3,length(t));
wm = 0; % white noise
vm = 0; % 

for i=1:length(t)-1
    Xm(:,i+1) = Am*Xm(:,i) + Bm*Xr(:,i) + wm;
    Xy(:,i)   = Cm*Xm(:,i) + vm;
end

% Motion Equations 
Xp = zeros(2,length(t));

for i=1:length(t)-1
    Bp = [Ts*cos(Xy(2,i))  0 0;
          Ts*sin(Xy(2,i))  0 0];
    Xp(:,i+1) = Xp(:,i) + Bp*Xy(:,i);
end
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% Perform extend Kalman filtering
%--------------------------------------------------------------------------

for i = 2:length(t)-1
    
    % Update true state
    x(:,i) = [Xr(:,i); Xm(:,i); Xy(:,i); Xp(:,i) ]; %+ sqrt(Q)*randn(nA,1);

    % Update measurements
    z(:,i) = H*x(:,i) + sqrt(R)*randn(nC,1);
    
    % Update state Jacobian
    Ai = StateEquation(sysd,Ts,x_aposteriori(14:16,i-1));
    %Ai = StateEquation(Ts,Xy(:,i-1));
    
    % Update apriori estimate
    x_apriori(:,i) = Ai*x_aposteriori(:,i-1);

    % Update aprioiri error covariance estimate
    P_apriori(:,:,i) = Ai*P_aposteriori(:,:,i-1)*Ai' + Q;

    % Update Kalman gain
    K(:,:,i) = P_apriori(:,:,i)*H' / (H*P_apriori(:,:,i)*H'+R);

    % Update aposteriori state estimate
    x_aposteriori(:,i) = x_apriori(:,i) + K(:,:,i) * (z(:,i) - H*x_apriori(:,i));

    % Update aposteriori error covariance estimate
    %P_aposteriori(:,:,i) = (eye(nA) - K(:,:,i)*H) * P_apriori(:,:,i);
    P_aposteriori(:,:,i) = (eye(nA) - K(:,:,i)*H) * P_apriori(:,:,i) * (eye(nA) - K(:,:,i)*H)' + K(:,:,i) *R*K(:,:,i)';
end
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% Plot Results
%--------------------------------------------------------------------------
%plotting;
testplotting;