//***************************************************************************
// Copyright 2007-2014 Universidade do Porto - Faculdade de Engenharia      *
// Laboratório de Sistemas e Tecnologia Subaquática (LSTS)                  *
//***************************************************************************
// This file is part of DUNE: Unified Navigation Environment.               *
//                                                                          *
// Commercial Licence Usage                                                 *
// Licencees holding valid commercial DUNE licences may use this file in    *
// accordance with the commercial licence agreement provided with the       *
// Software or, alternatively, in accordance with the terms contained in a  *
// written agreement between you and Universidade do Porto. For licensing   *
// terms, conditions, and further information contact lsts@fe.up.pt.        *
//                                                                          *
// European Union Public Licence - EUPL v.1.1 Usage                         *
// Alternatively, this file may be used under the terms of the EUPL,        *
// Version 1.1 only (the "Licence"), appearing in the file LICENCE.md       *
// included in the packaging of this file. You may not use this work        *
// except in compliance with the Licence. Unless required by applicable     *
// law or agreed to in writing, software distributed under the Licence is   *
// distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF     *
// ANY KIND, either express or implied. See the Licence for the specific    *
// language governing permissions and limitations at                        *
// https://www.lsts.pt/dune/licence.                                        *
//***************************************************************************
// Author: Kristian Klausen                                                 *
//***************************************************************************

// ISO C++ 98 headers.
#include <cmath>
#include <string>
#include <vector>

// DUNE headers.
#include <DUNE/DUNE.hpp>

namespace Control
{
  namespace UAV
  {
    namespace MyDerivedGuidanceController
    {
      using DUNE_NAMESPACES;
      using std::sin;
      using std::cos;
      using std::tan;

      //! Controllable loops
      //! Which loop makes this task start!
      static const uint32_t c_controllable = IMC::CL_GUIDANCE;
      //! Required loops
      //! Aka, which loops to start when this task starts.
      static const uint32_t c_required = IMC::CL_ROLL | CL_PITCH | CL_ALTITUDE | CL_YAW | CL_FORCE | CL_VELOCITY;


      struct Arguments
      {
        float roll;
        float pitch;
        float yaw;
        float z_alt;
        float z_thrust;
        float vel_x;
        float vel_y;
        int mode_rp;
        int mode_yaw;
        int mode_z;
      };

      struct Task: public DUNE::Control::BasicGuidanceController
      {
        float m_roll;
        float m_pitch;
        float m_yaw;
        float m_z_alt;
        float m_z_thrust;
        float m_vel_x;
        float m_vel_y;
        int m_mode_rp;
        int m_mode_yaw;
        int m_mode_z;

        //! Task Arguments
        Arguments m_args;

        Task(const std::string& name, Tasks::Context& ctx):
          DUNE::Control::BasicGuidanceController(name, ctx, c_controllable, c_required),
          m_roll(0),
          m_pitch(0),
          m_yaw(0),
          m_z_alt(0),
          m_z_thrust(0),
          m_vel_x(0),
          m_vel_y(0),
          m_mode_rp(0),
          m_mode_yaw(0),
          m_mode_z(0)
        {

          param("Desired Roll", m_args.roll)
          .defaultValue("0")
          .units(Units::Degree)
          .description("Roll angle");

          param("Desired Pitch", m_args.pitch)
          .defaultValue("0")
          .units(Units::Degree)
          .description("Pitch angle");

          param("Desired Yaw", m_args.yaw)
          .defaultValue("0")
          .units(Units::Degree)
          .description("Yaw angle");

          param("Desired Z - ALT", m_args.z_alt)
          .defaultValue("20")
          .units(Units::Meter)
          .description("Desired Z altitude");

          param("Desired Z - THRUST", m_args.z_thrust)
          .defaultValue("0")
          .description("Desired Z thrust: Range 0~100");

          param("Desired Velocity X", m_args.vel_x)
          .defaultValue("0")
          .units(Units::MeterPerSecond)
          .description("Desired Velocity X in body");

          param("Desired Velocity Y", m_args.vel_y)
          .defaultValue("0")
          .units(Units::MeterPerSecond)
          .description("Desired Velocity Y in body");

          param("Mode RP", m_args.mode_rp)
          .defaultValue("2")
          .description("Roll-pitch mode: RP: 1, XY: 2");

          param("Mode Yaw", m_args.mode_yaw)
          .defaultValue("0")
          .description("Yaw mode: control: 1");

          param("Mode Z", m_args.mode_z)
          .defaultValue("1")
          .description("Z mode: ALT: 1, THRUST: 2");


          inf("HURRAY!");
        }

        //! Initialize resources
        void
        onResourceInitialization(void)
        {
          BasicGuidanceController::onResourceInitialization();
        }

        //! Acquire resources
        void
        onResourceAcquisition(void)
        {
          BasicGuidanceController::onResourceAcquisition();
        }

        //! Release Resources
        void
        onResourceRelease(void)
        {
          BasicGuidanceController::onResourceRelease();
        }

        //! On activation
        void
        onAutopilotActivation(void)
        {
          // Do whatever
          inf("Pilot activated!");
        }

        //! On deactivation
        void
        onAutopilotDeactivation(void)
        {
          // Do whatever
        }

        //! Update internal parameters.
        void
        onUpdateParameters(void)
        {
          reset();
          initializePIDs();

          if (paramChanged(m_args.roll))
            m_roll = Angles::radians(m_args.roll);
          if (paramChanged(m_args.pitch))
            m_pitch = Angles::radians(m_args.pitch);
          if (paramChanged(m_args.yaw))
            m_yaw = Angles::radians(m_args.yaw);
          if (paramChanged(m_args.z_alt))
            m_z_alt = m_args.z_alt;
          if (paramChanged(m_args.z_thrust))
            m_z_thrust = m_args.z_thrust/100;
          if (paramChanged(m_args.vel_x))
            m_vel_x = m_args.vel_x;
          if (paramChanged(m_args.vel_y))
            m_vel_y = m_args.vel_y;
          if (paramChanged(m_args.mode_rp))
            m_mode_rp = m_args.mode_rp;
          if (paramChanged(m_args.mode_yaw))
            m_mode_yaw = m_args.mode_yaw;
          if (paramChanged(m_args.mode_z))
            m_mode_z = m_args.mode_z;
        }

        //! Initialize PID related variables
        void
        initializePIDs(void)
        {

        }

        //! Member variable reset function
        void
        reset(void)
        {
          BasicGuidanceController::reset();
        }

        //! Reserve entities for messages
        void
        onEntityReservation(void)
        {
          // Seldom needed.
        }

        //! Computes control values when receiving EstimatedState
        //! @param[in] timestep time interval to use in pid controller
        //! @param[in] msg pointer to EstimatedState message
        void
        onEstimatedState(const double timestep, const IMC::EstimatedState* msg)
        {
          // Do my control
          inf("I am in Control!");

          IMC::DesiredVelocity vel;
          vel.flags |= IMC::DesiredVelocity::FL_SWAY | IMC::DesiredVelocity::FL_SURGE;
          vel.u = m_vel_x;
          vel.v = m_vel_y;

          IMC::DesiredYaw yaw;
          yaw.value = m_yaw;

          IMC::DesiredPitch pitch;
          IMC::DesiredRoll roll;

          pitch.value = m_pitch;
          roll.value  = m_roll;

          IMC::DesiredZ alt;
          IMC::DesiredControl ctrl;
          if (m_mode_z == 1) {
            alt.z_units = Z_ALTITUDE;
            alt.value = m_z_alt;
          } else if (m_mode_z == 2) {
            ctrl.flags |= IMC::DesiredControl::FL_Z;
            ctrl.z = m_z_thrust;
          }

          if (m_mode_rp == 1) {
            dispatch(roll);
            dispatch(pitch);
          } else if (m_mode_rp == 2) {
            dispatch(vel);
          }

          if (m_mode_z == 1) {
            dispatch(alt);
          } else if (m_mode_z == 2) {
            dispatch(ctrl);
          }

          if (m_mode_yaw) {
            dispatch(yaw);
          }
        }
      };
    }
  }
}

DUNE_TASK
